#!/usr/bin/python
# -*- coding: utf-8 -*-
#------------------------------------------------------------------------
# 魚眼レンズを使用し,動体検知後,顔およびネコを検出し,検出した画像を記録する
#
# 参考URL:
#   https://ensekitt.hatenablog.com/entry/2018/06/11/200000
#
# インストール方法:
#   sudo apt-get update
#   sudo apt-get -y upgrade
#   sudo apt-get install -y libopencv-dev
#   sudo apt-get install -y python-opencv
#   wget https://raw.githubusercontent.com/opencv/opencv/master/data/haarcascades/haarcascade_frontalcatface.xml
#   wget https://raw.githubusercontent.com/opencv/opencv/master/data/haarcascades/haarcascade_frontalface_default.xml
#------------------------------------------------------------------------
import cv2
import sys
import numpy as np
import time,datetime
import math

debug = False

DIAMETER=800        #直径
RADIUS=DIAMETER//2  #半径
WIDTH=DIAMETER
HEIGHT=DIAMETER

#DEWARP_RESIZE=True
DEWARP_RESIZE=False
ANGLE=1
if (DEWARP_RESIZE is True):
    ARC_LENGTH=2
else:
    ARC_LENGTH=4

DIFF_COUNT=3000
MAX_DIFF_COUNT=20000

last_frame = None

face_cascade_filename = "haarcascade_frontalface_default.xml"
cat_cascade_filename="haarcascade_frontalcatface.xml"

#------------------------------------------------------------------------
# 指定座標の角度を求める  
# 入力
#   cx,cy --- 中心点
#   x,y ----- 指定座標
# 出力
#   k ------- 角度(0〜359度)
#------------------------------------------------------------------------
def get_degrees(cx, cy, x, y):
    px,py = (float)(x - cx), (float)(y - cy)
    if (x == cx):
        k = 0 if (y > cx) else 180
    else:
        n = int(math.degrees(math.atan(py/px)))
        k = (270 if (x >= cx) else 90) + n
    return k

#------------------------------------------------------------------------
# x,y,w,hで成形される矩形の各頂点への角度を求め,その最小角と最大角を求める
# 入力
#   cx,cy --- 中心点
#   x,y ----- 左上の座標
#   w,h ----- 幅,高さ
# 出力
#   min,max - 最小角,最大角
#------------------------------------------------------------------------
def get_range(cx, cy, x, y, w, h):
    n = np.zeros(4)
    n[0] = get_degrees(cx, cy, x  , y  )
    n[1] = get_degrees(cx, cy, x+w, y  )
    n[2] = get_degrees(cx, cy, x  , y+h)
    n[3] = get_degrees(cx, cy, x+w, y+h)
    min = np.amin(n)
    max = np.amax(n)
    return int(min),int(max)

#------------------------------------------------------------------------
# 顔を検出する
# 入力
#   img   --- 検出対象画像
# 出力
#   検出時 ---- True,検出した箇所を矩形で示した画像
#   未検出時 -- False
#------------------------------------------------------------------------
def detect_face(img):
    global face_cascade

    face_img = img.copy()
    face_rects = face_cascade.detectMultiScale(face_img)
    for (x,y,w,h) in face_rects:
        cv2.rectangle(face_img, (x,y), (x+w,y+h), (255,0,255), 1)
        return True,face_img
    return False,face_img

#------------------------------------------------------------------------
# ネコを検出する
# 入力
#   img   --- 検出対象画像
# 出力
#   検出時 ---- True,検出した箇所を矩形で示した画像
#   未検出時 -- False
#------------------------------------------------------------------------
def detect_cat(img):
    global cat_cascade

    cat_img = img.copy()
    cat_rects = cat_cascade.detectMultiScale(img, 1.1, 3)
    for (x, y, w, h) in cat_rects:
        cv2.rectangle(cat_img, (x,y), (x+w,y+h), (255,0,255), 1)
        return True,cat_img
    return False,cat_img

#------------------------------------------------------------------------
# 画像の動きを検出し,動き量,動きのある箇所を取得した画像,
# その箇所(矩形)の最小角度,最大角度を返す
# 入力
#   frame --- 入力画像
# 出力
#   動き量,動きのある箇所を取得した画像,最小角度,最大角度
# 備考
#   https://ensekitt.hatenablog.com/entry/2018/06/11/200000
#------------------------------------------------------------------------
def detect_motion(frame):
    global last_frame

    gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    if (last_frame is None):
        last_frame = gray.copy().astype('float')
        return 0,last_frame,0,0

    cv2.accumulateWeighted(gray, last_frame, 0.5)
    mdframe = cv2.absdiff(gray, cv2.convertScaleAbs(last_frame))
    thresh = cv2.threshold(mdframe, 3, 255, cv2.THRESH_BINARY)[1]
    image, contours, hierarchy = cv2.findContours(thresh.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    max = 0
    target = contours[0]
    for n in contours:
        area = cv2.contourArea(n)
        if ((max < area) and (area < MAX_DIFF_COUNT) and (area > DIFF_COUNT)):
            max = area;
            target = n

    if (max <= DIFF_COUNT):
        return 0,mdframe,0,0

    x,y,w,h = cv2.boundingRect(target)
    height,width = gray.shape[:2]
    height,width = height // 2, width // 2
    mdframe = cv2.rectangle(mdframe,(x,y),(x+w,y+h),(255,255,255),2)
    mdframe = cv2.circle(mdframe, (width,height), width, (255, 255, 255), thickness=1)

    amin,amax = get_range(width, height, x, y, w, h)
    if (amax - amin > 180):
        amin,amax = amax,amin

    return int(max),mdframe,amin,amax

#------------------------------------------------------------------------
# 魚眼画像を指定した角度分だけ変形する
# 入力
#   img_src --- 入力画像
#   t --------- 角度(何度づつ取り出すかを指定する)
#   w --------- 弧
#   sp -------- 開始角度
#   ep -------- 終了角度
# 出力
#   魚眼画像を変形した画像
#------------------------------------------------------------------------
def dewarp(img_src, t, w, sp, ep):
    if (DEWARP_RESIZE is True):
        img_src = cv2.resize(img_src, (img_src.shape[1]//2,img_src.shape[0]//2))
    center = tuple(np.array([img_src.shape[1] * 0.5, img_src.shape[0] * 0.5]))
    x,y = center
    cx,cy = int(x), int(y)
 
    height,width = img_src.shape[:2]
    size = tuple(np.array([width, height]))
 
    scale = 1.0

    w = ARC_LENGTH
    x = int(cx - w/2)
    y = int(cy + cy/2)
    h = int(height-y)
 
    if (sp > ep):
        for angle in range(sp, ep, -t):
            rotation_matrix = cv2.getRotationMatrix2D(center, angle, scale)
            img_rot = cv2.warpAffine(img_src, rotation_matrix, size, flags=cv2.INTER_CUBIC)
            clipped = img_rot[y:y+h, x:x+w : ]
            img_dewarp = clipped if (angle == sp) else cv2.hconcat([img_dewarp, clipped])
    else:
        for angle in range(sp, 0, -t):
            rotation_matrix = cv2.getRotationMatrix2D(center, angle, scale)
            img_rot = cv2.warpAffine(img_src, rotation_matrix, size, flags=cv2.INTER_CUBIC)
            clipped = img_rot[y:y+h, x:x+w : ]
            img_dewarp = clipped if (angle == sp) else cv2.hconcat([img_dewarp, clipped])
        for angle in range(360, ep, -t):
            rotation_matrix = cv2.getRotationMatrix2D(center, angle, scale)
            img_rot = cv2.warpAffine(img_src, rotation_matrix, size, flags=cv2.INTER_CUBIC)
            clipped = img_rot[y:y+h, x:x+w : ]
            img_dewarp = cv2.hconcat([img_dewarp, clipped])

    return img_dewarp

#------------------------------------------------------------------------
# 魚眼画像を展開する
# 入力
#   魚眼画像
# 出力
#   展開した画像
#------------------------------------------------------------------------
def spread(img_src):
    center = tuple(np.array([img_src.shape[1] * 0.5, img_src.shape[0] * 0.5]))
    x,y = center
    cx,cy = int(x), int(y)
 
    height,width = img_src.shape[:2]
    size = tuple(np.array([width, height]))
 
    scale = 1.0

    w = ARC_LENGTH
    x = int(cx - w/2)
    y = int(cy + cy/2)
    h = int(height-y)
 
    for angle in range(360, -20, -ANGLE):
        rotation_matrix = cv2.getRotationMatrix2D(center, angle, scale)
        img_rot = cv2.warpAffine(img_src, rotation_matrix, size, flags=cv2.INTER_CUBIC)
        clipped = img_rot[y:y+h, x:x+w : ]
        if (angle == 360):
            img_spread = clipped
        else:
            img_spread = cv2.hconcat([img_spread, clipped])
    return img_spread

#------------------------------------------------------------------------
#
#------------------------------------------------------------------------
def init():
    global face_cascade, cat_cascade

    if (debug == True):
        r = RADIUS
        L = int(2*3.14*r)        #円周
        l = int(L/360*ANGLE)     #弧
        print("ANGLE=", ANGLE, " ARC_LENGTH=", ARC_LENGTH, " L=", L, " l=", l)

    face_cascade = cv2.CascadeClassifier(face_cascade_filename)
    if (face_cascade.empty()):
        print("cannnot load face cascade file")
        sys.exit(-1)
    print("loaded face cascade file:", face_cascade_filename)

    cat_cascade = cv2.CascadeClassifier(cat_cascade_filename)
    if (cat_cascade.empty()):
        print("cannnot load cat  cascade file")
        sys.exit(-1)
    print("loaded cat  cascade file:", cat_cascade_filename)

#------------------------------------------------------------------------
#
#------------------------------------------------------------------------
def write_image(src,face):
    src_fname  = "detect_{0:%Y%m%d-%H%M%S}.jpg".format(datetime.datetime.now())
    #### face_fname = "detect_{0:%Y%m%d-%H%M%S}_face.jpg".format(datetime.datetime.now())
    print(src_fname)
    cv2.imwrite(src_fname, src)
    #### cv2.imwrite(face_fname, face)

#------------------------------------------------------------------------
#
#------------------------------------------------------------------------
def monitor():
    cap = cv2.VideoCapture(0)
    cap.set(cv2.CAP_PROP_FRAME_WIDTH,  DIAMETER) # カメラ画像の横幅
    cap.set(cv2.CAP_PROP_FRAME_HEIGHT, DIAMETER) # カメラ画像の縦幅

    #### 変形画像表示用windowを作成しておく(最初の動き未検出時のクリア処理のため)
    dewarp_stab = np.zeros((1, 1, 3), np.uint8)
    dewarp_stab = cv2.cvtColor(dewarp_stab, cv2.COLOR_BGR2GRAY)
    cv2.imshow("dewarp", dewarp_stab)
    cv2.moveWindow("dewarp", 50, RADIUS+50+50)

    print("type 'Esc' or 'q' to quit.")

    interval = 1
    wstime = time.time()
    while (cap.isOpened()):

        key = cv2.waitKey(1) & 0xFF
        if (key == 27 or key == ord('q')):
            break

        #### interval 間,空読みする
        wetime = time.time()
        if (wetime - wstime < interval):
             ret, img_src = cap.read()
             continue
        wstime = time.time()

        #### 入力画像を縮小して表示する
        img_src_resize = cv2.resize(img_src, (RADIUS,RADIUS))
        cv2.imshow("image", img_src_resize)
        cv2.moveWindow("image", 50, 50)

        st = time.time()  #### intervalを求める為のもの

        #### 動きを検出する
        n,diff,amin,amax = detect_motion(img_src)
        if (n == 0): #### 画像に動きがないときは
            #### 動き検出画像と変形画像をクリアする
            diff_resize = cv2.resize(diff, (RADIUS,RADIUS))
            diff_resize = cv2.bitwise_not(diff_resize)
            cv2.imshow("diff", diff_resize)
            cv2.moveWindow("diff", 50+RADIUS+10, 50)

            cv2.imshow("dewarp", dewarp_stab)
            cv2.destroyWindow("dewarp")
        else: #### 画像に動きがあるときは
            et = time.time()
            #### 動き検出画像を縮小し,反転して表示する
            diff_resize = cv2.resize(diff, (RADIUS,RADIUS))
            diff_resize = cv2.bitwise_not(diff_resize)
            cv2.imshow("diff", diff_resize)
            cv2.moveWindow("diff", 50+RADIUS+10, 50)

            #### 変形画像を求める
            gray = cv2.cvtColor(img_src, cv2.COLOR_BGR2GRAY)
            img_dewarp = dewarp(gray, ANGLE, ARC_LENGTH,amax,amin)
            et = time.time()
            if (debug is True): print("dewarp w=", img_dewarp.shape[1], " time=", et-st)
            interval = et - st + 1
         
            #### 変形画像を表示する
            cv2.imshow("dewarp", img_dewarp)
            cv2.moveWindow("dewarp", 50, RADIUS+50+50)

            #### 顔を検出する(検出時は、画像を書き出す)
            status,face_img = detect_face(img_dewarp)
            if (status is True):
                write_image(img_src,face_img)

            #### ネコを検出する(検出時は、画像を書き出す)
            status,cat_img = detect_cat(img_dewarp)
            if (status is True):
                write_image(img_src,cat_img)

    cv2.destroyAllWindows()

#------------------------------------------------------------------------
#
#------------------------------------------------------------------------
def analysis(fname):
    if (fname.endswith(".jpg") != True):
        help()
        sys.exit(1)
    if (fname.startswith("detect_") != True):
        help()
        sys.exit(1)
    if (fname.startswith("analysis_") == True):
        print("already analyzed.")
        sys.exit(0)
    img = cv2.imread(fname)

    #### 魚眼画像を展開する
    img_dewarp = spread(img)
    
    #### 顔を検出する(検出時は、画像を書き出す)
    status,img_dewarp = detect_face(img_dewarp)

    #### ネコを検出する(検出時は、画像を書き出す)
    status,img_dewarp = detect_cat(img_dewarp)

    cv2.imshow("analysis", img_dewarp)
    cv2.moveWindow("analysis", 50, 50)

    print("type 'Esc' or 'q' to quit.")
    print("type 'p' to write and quit.")
    while (True):
        key = cv2.waitKey(0) & 0xFF
        if (key == 27 or key == ord('q')):
            break
        elif (key == ord('p')):
            fname = fname.replace("detect", "analysis")
            cv2.imwrite(fname, img_dewarp)
            break

#------------------------------------------------------------------------
#
#------------------------------------------------------------------------
def help():
    print("usage")
    print(" $ python detect_cat.py")
    print("     monitor and save detected image to detect_YYYYMMDD-hhmmss.jpg")
    print(" $ python detect_cat.py --help (or -h)")
    print("     display this message")
    print(" $ python detect_cat.py detect_YYYYMMDD-hhmmss.jpg")
    print("     analysis detect_YYYYMMDD-hhmmss.jpg to analysis_YYYYMMDD-hhmmss.jpg")

#------------------------------------------------------------------------
#
#------------------------------------------------------------------------
def main():
    args = sys.argv
    argc = len(args)

    if (argc == 1):
        monitor()  
        sys.exit(0)
    if ("-h" in args[1]):
        help()
        sys.exit(0)
    analysis(args[1])

#------------------------------------------------------------------------
#
#------------------------------------------------------------------------
if __name__ == '__main__':
    init()
    main()
